==> makeen-expressionengine-0.0.51.tgz: package/src/expressionData.ts <==
﻿import { Stack } from "./utils/stack";
import List from 'typescript-dotnet-commonjs/System/Collections/List';
import { makeenBindingContext } from './bindingContext';
import { indexerPart } from "./compiler/parts/indexerPart";
import { EventHandler, ExpressionComponentChangedHandler, PropertyChangedEventHandler, NotifyCollectionChangedEventHandler } from "./utils/events/eventHandlers";
import { INotifyPropertyChanged } from "./utils/events/INotifyPropertyChanged";
import { expressionCompiler } from "./compiler/expressionCompiler";
import { expressionPart } from "./compiler/parts/expressionPart";
import { propertyPart } from "./compiler/parts/propertyPart";
import { INotifyCollectionChanged } from "./utils/events/INotifyCollectionChanged";
import { getHashCode } from "typescript-dotnet-commonjs/System/Text/Utility";
import './utils/stringUtil';
export default class expressionData {

    private boundCollections: List<HandlerRecord> = new List<HandlerRecord>();

    private boundObjects: List<HandlerRecord> = new List<HandlerRecord>();

    private subExpressions: List<expressionData> = new List<expressionData>();

    private pageContext!: makeenBindingContext;

    private globalContext!: makeenBindingContext;

    public parent!: expressionData;

    public expression!: string

    public contexts: Stack<Object> = new Stack<object>();

    public convertType: boolean = false;

    public monitorExpression: boolean = false;
    public cacheEnabled: boolean = false;

    public Changed!: ExpressionComponentChangedHandler;

    public legacyMode: boolean = false;

    public isAlreadyLegacy: Boolean = false;

    public controlDestroyed!: EventHandler<Object>;

    public OnControlDestroyed(control: Object) {
        if ((this.controlDestroyed != null)) {
            this.controlDestroyed.invoke(null, control);
        }
    }


    public constructor(expression: string, pageContext?: Object, globalContext?: Object, legacyMode: boolean = true, convertType: boolean = true) {
        this.constructorInit(expression, pageContext, globalContext, legacyMode, convertType, false)
        this.cacheEnabled = true;
    }

    private constructorInit(expression: string, pageContext?: Object, globalContext?: Object, legacyMode?: boolean, convertType: boolean = true, isAlreadyLegacy: boolean = false) {
        this.expression = expression;
        this.convertType = convertType;
        this.globalContext = this.globalContext;
        this.pageContext = this.pageContext;
        this.isAlreadyLegacy = isAlreadyLegacy;
        this.legacyMode = legacyMode as boolean;
        this.cacheEnabled = true;
        this.contexts = new Stack<Object>();
        if ((this.globalContext != null)) {
            this.contexts.push(this.globalContext);
        }

        if ((this.pageContext != null)) {
            this.contexts.push(this.pageContext);
        }

        this.toLegacyIfNeeded();
        // ControlDestroyed += ExpressionData_ControlDestroyed;
    }

    private privateConstructor(expression: string, contexts: Stack<Object>, pageContext: makeenBindingContext, globalContext: makeenBindingContext, legacyMode: boolean, convertType: boolean = true, isAlreadyLegacy: Boolean = false) {
        this.expression = expression;

        this.convertType = convertType;
        this.globalContext = globalContext;
        this.pageContext = pageContext;
        this.contexts = contexts.clone();
        //this.contexts.push(contexts);
        this.isAlreadyLegacy = isAlreadyLegacy;
        this.legacyMode = legacyMode;
        this.cacheEnabled = true;
        this.toLegacyIfNeeded();
        return this;
        // ControlDestroyed += ExpressionData_ControlDestroyed;
    }

    ExpressionData_ControlDestroyed(sender: Object, boundObject: Object) {
        let removables: List<HandlerRecord> = new List<HandlerRecord>();
        this.boundCollections.forEach((item) => {
            if (item.checkBound(boundObject)) {
                removables.add(item);
                item.unbind();
            }
        })
        this.boundObjects.forEach((item) => {
            if (item.checkBound(boundObject)) {
                removables.add(item);
                item.unbind();
            }
        })
        removables.forEach((item) => {
            this.boundCollections.remove(item);
            this.boundObjects.remove(item);
        });
    }

    public toSubEx(currentContext: Object | null = null): expressionData {
        return this.toSubExInit(this.expression, currentContext);
    }

    public toSubExInit(newExpression: string, currentContext: Object | null = null): expressionData {
        let subExpressionData: expressionData = this.privateConstructor(newExpression, this.contexts, this.pageContext, this.globalContext, this.legacyMode, this.convertType, this.isAlreadyLegacy);
        if (((currentContext != null)
            && ((currentContext != this.pageContext)
                && ((currentContext != this.globalContext)
                    && !subExpressionData.contexts.contains(currentContext))))) {
            subExpressionData.contexts.push(currentContext);
        }

        if (this.parent != null) {
            subExpressionData.parent = this.parent;
        }
        else {
            subExpressionData = this;
        }
        subExpressionData.cacheEnabled = this.cacheEnabled;
        this.subExpressions.add(subExpressionData);
        return subExpressionData;
    }

    subExpressionData_Changed(context: Object, propertyName: string, expression: string) {
        this.NotifyChanged(context, context, expression);
    }

    public NotifyChanged(context: Object, subContext: Object, propertyName: string) {
        if ((this.Changed != null)) {
            this.Changed.invoke(context, subContext, propertyName, this.expression);
        }

        if ((this.parent != null)) {
            this.parent.NotifyChanged(context, subContext, propertyName);
        }

    }

    private isBound(obj: unknown): boolean {
        if (obj instanceof INotifyCollectionChanged) {
            this.boundCollections.forEach((item) => {
                if (item.checkBound(obj)) {
                    return true;
                }
            })
        } else if (obj instanceof INotifyPropertyChanged) {
            this.boundObjects.forEach((item) => {
                if (item.checkBound(obj)) {
                    return true;
                }
            });

            return false;
        }
        return false;
    }

    public monitor(propertyName: string, bindableResult: unknown) {
        if (bindableResult instanceof INotifyPropertyChanged) {
            if (!this.isBound(bindableResult)) {
                let handler: PropertyChangedEventHandler = new PropertyChangedEventHandler();
                handler.add((sender, e) => {
                    if (propertyName != null && e.propertyName != null && (propertyName.equals(e.PropertyName) || propertyName.replace("[\"", "[").replace("\"]", "]").equals(e.PropertyName))) {
                        this.NotifyChanged(bindableResult, sender, e.propertyName);
                    }
                    else if (e.PropertyName != null && e.PropertyName.StartsWith(propertyName + "[")) {
                        this.NotifyChanged(bindableResult, sender, e.propertyName);
                    }
                });

                bindableResult.propertyChanged = handler;
                this.addHandler(bindableResult, handler);
            }
        }
        else if (bindableResult instanceof INotifyCollectionChanged) {
            let handler: NotifyCollectionChangedEventHandler = new NotifyCollectionChangedEventHandler();
            handler.add((sender, e) => {
                if (propertyName != null && e.propertyName != null && (propertyName.equals(e.PropertyName) || propertyName.replace("[\"", "[").replace("\"]", "]").equals(e.PropertyName))) {
                    this.NotifyChanged(bindableResult, sender, e.propertyName);
                }
                else if (e.PropertyName != null && e.PropertyName.StartsWith(propertyName + "[")) {
                    this.NotifyChanged(bindableResult, sender, e.propertyName);
                }
            });
            bindableResult.CollectionChanged = handler;
            this.addHandler(bindableResult, handler);
        }

    }

    public addHandler(bindableResult: unknown, handler: PropertyChangedEventHandler) {
        if (bindableResult instanceof INotifyPropertyChanged) {

            let objectHandler: ObjectHandlerRecord = new ObjectHandlerRecord();
            objectHandler.Handler = new PropertyChangedEventHandler;
            objectHandler.Object = bindableResult;

            this.boundObjects.add(objectHandler);
        }
        else if (bindableResult instanceof INotifyCollectionChanged) {
            let boundCollections: CollectionHandlerRecord = new CollectionHandlerRecord();
            boundCollections.Handler = handler;
            boundCollections.Object = bindableResult;
            this.boundCollections.add(boundCollections);
        }
    }

    public UnBindHandlers() {
        while ((this.boundObjects.count > 0)) {
            (new List<HandlerRecord>(this.boundObjects)).forEach((item) => {

                this.boundObjects.clear();
            });
        }

        while ((this.boundCollections.count > 0)) {
            (new List<HandlerRecord>(this.boundCollections)).forEach((item) => {
                item.unbind();
            });

            this.boundCollections.clear();
        }

    }

    public /* override */ equals(obj: Object): boolean {
        if ((obj instanceof expressionData)) {
            let ed: expressionData = (<expressionData>(obj));
            return (this.expression == ed.expression);
        }

        return false;
    }

    public getHashCode(): number {
        let hash: number = 13;
        //  hash = (hash * 7) + (this.Binding != null ? this.Binding.GetHashCode() : 0);
        hash = ((hash * 7) + getHashCode(this.expression));
        return hash;
    }

    private toLegacyIfNeeded() {
        if ((!this.isAlreadyLegacy
            && this.legacyMode)) {
            this.isAlreadyLegacy = true;
            this.expression = expressionCompiler.ProcessLegacy(this.expression);
        }

    }

    public Dispose() {
        this.Destroy();
    }

    public Destroy() {
        this.contexts.clear();
        // this.resultCache.clear();
        this.UnBindHandlers();
        // this.DestroySubParts();
        // ControlDestroyed = (ControlDestroyed - this.ExpressionData_ControlDestroyed);
        // this.globalContext = null;
        // this.pageContext = null;
    }

    // public DestroySubParts() {
    //     while ((this.subExpressions.count > 0)) {
    //         (new List<expressionData>(this.subExpressions)).forEach((item) => {
    //             let subExpression: expressionData;
    //             if (item.tryGetTarget(subExpression)) {
    //                 subExpression.Destroy();
    //             }
    //         });

    //         this.subExpressions.clear();
    //     }

    // }

    public toString(): string {
        return this.expression;
    }

    // constructor() {
    //     this.Dispose();
    // }

    public monitorIfRequired(item: expressionPart| undefined, currentContext: Object) {
        let propertyName: string = "";
        if (this.monitorExpression) {
            if ((currentContext instanceof INotifyCollectionChanged)) {
                this.monitor("", currentContext);
            }

            if ((currentContext instanceof INotifyPropertyChanged && (item != null))) {
                if ((item instanceof propertyPart)) {
                    propertyName = (<propertyPart>(item)).propertyName;
                }
                else if ((item instanceof indexerPart)) {
                    propertyName = item.toString();
                }

                if ((propertyName != null)) {
                    this.monitor(propertyName, (<INotifyPropertyChanged>(currentContext)));
                }

            }

        }

    }
}

abstract class HandlerRecord {

    public abstract unbind(): void;

    public abstract checkBound(obj: Object): boolean;
}

class ObjectHandlerRecord extends HandlerRecord {

    public Object: INotifyPropertyChanged | undefined;

    public Handler: PropertyChangedEventHandler | undefined;

    public unbind() {
        // let obj: INotifyPropertyChanged;
        // let handler: PropertyChangedEventHandler;
        // if (this.Object.TryGetTarget(/* out */obj)) {
        //     if (this.Handler.TryGetTarget(/* out */handler)) {
        //         obj.PropertyChanged = (obj.PropertyChanged - handler);
        //     }

        // }

    }

    public checkBound(obj: Object): boolean {
        let tryObj: INotifyPropertyChanged;
        if (this.Object instanceof INotifyPropertyChanged) {
            return true;
        }

        return false;
    }
}

class CollectionHandlerRecord extends HandlerRecord {

    public Object: INotifyCollectionChanged | undefined;

    public Handler: NotifyCollectionChangedEventHandler | undefined;

    public unbind() {
        // let obj: INotifyCollectionChanged;
        // let handler: NotifyCollectionChangedEventHandler;
        // if (this.Object.TryGetTarget(/* out */obj)) {
        //     if (this.Handler.TryGetTarget(/* out */handler)) {
        //         obj.CollectionChanged = (obj.CollectionChanged - handler);
        //     }

        // }

    }

    public checkBound(obj: Object): boolean {
        let tryObj: INotifyCollectionChanged;
        if (this.Object instanceof INotifyCollectionChanged) {
            return true;
        }

        return false;
    }
}

==> makeen-expressionengine-0.0.63.tgz: package/src/expressionData.ts <==
﻿import { Stack } from "./utils/stack";
import List from 'typescript-dotnet-commonjs/System/Collections/List';
import { makeenBindingContext } from './bindingContext';
import { expressionCompiler } from "./compiler/expressionCompiler";
// import { getHashCode } from "typescript-dotnet-commonjs/System/Text/Utility";
import './utils/stringUtil';

export default class expressionData {

    private pageContext!: makeenBindingContext;

    private globalContext!: makeenBindingContext;

    parent!: expressionData;

    private subExpressions: List<expressionData> = new List<expressionData>();

    expression!: string

    contexts: Stack<Object> = new Stack<object>();

    convertType: boolean = false;

    monitorExpression: boolean = false;
    cacheEnabled: boolean = false;

    legacyMode: boolean = false;

    isAlreadyLegacy: Boolean = false;

    dtFormat: string | undefined = undefined;

    constructor(expression: string, pageContext?: Object, globalContext?: Object, legacyMode: boolean = true, convertType: boolean = true) {
        this.constructorInit(expression, pageContext, globalContext, legacyMode, convertType, false)
        this.cacheEnabled = true;
    }

    private constructorInit(expression: string, pageContext?: Object, globalContext?: Object, legacyMode?: boolean, convertType: boolean = true, isAlreadyLegacy: boolean = false) {
        this.expression = expression;
        this.convertType = convertType;
        this.globalContext = this.globalContext;
        this.pageContext = this.pageContext;
        this.isAlreadyLegacy = isAlreadyLegacy;
        this.legacyMode = legacyMode as boolean;
        this.cacheEnabled = true;
        this.contexts = new Stack<Object>();
        if ((this.globalContext != null)) {
            this.contexts.push(this.globalContext);
        }

        if ((this.pageContext != null)) {
            this.contexts.push(this.pageContext);
        }

        this.toLegacyIfNeeded();
    }

    private privateConstructor(expression: string, contexts: Stack<Object>, pageContext: makeenBindingContext, globalContext: makeenBindingContext, legacyMode: boolean, convertType: boolean = true, isAlreadyLegacy: Boolean = false) {
        this.expression = expression;

        this.convertType = convertType;
        this.globalContext = globalContext;
        this.pageContext = pageContext;
        this.contexts = contexts.clone();

        this.isAlreadyLegacy = isAlreadyLegacy;
        this.legacyMode = legacyMode;
        this.cacheEnabled = true;
        this.toLegacyIfNeeded();
        return this;
    }

    toSubEx(currentContext: Object | null = null): expressionData {
        return this.toSubExInit(this.expression, currentContext);
    }

    toSubExInit(newExpression: string, currentContext: Object | null = null): expressionData {
        let subExpressionData: expressionData = this.privateConstructor(newExpression, this.contexts, this.pageContext, this.globalContext, this.legacyMode, this.convertType, this.isAlreadyLegacy);
        if (((currentContext != null)
            && ((currentContext != this.pageContext)
                && ((currentContext != this.globalContext)
                    && !subExpressionData.contexts.contains(currentContext))))) {
            subExpressionData.contexts.push(currentContext);
        }

        if (this.parent != null) {
            subExpressionData.parent = this.parent;
        }
        else {
            subExpressionData = this;
        }
        subExpressionData.cacheEnabled = this.cacheEnabled;
        this.subExpressions.add(subExpressionData);
        return subExpressionData;
    }

    /* override */ equals(obj: Object): boolean {
        if ((obj instanceof expressionData)) {
            let ed: expressionData = (<expressionData>(obj));
            return (this.expression == ed.expression);
        }

        return false;
    }

    // getHashCode(): number {
    //     let hash: number = 13;
    //     //  hash = (hash * 7) + (this.Binding != null ? this.Binding.GetHashCode() : 0);
    //     hash = ((hash * 7) + getHashCode(this.expression));
    //     return hash;
    // }

    private toLegacyIfNeeded() {
        if ((!this.isAlreadyLegacy
            && this.legacyMode)) {
            this.isAlreadyLegacy = true;
            this.expression = expressionCompiler.ProcessLegacy(this.expression);
        }

    }
    toString(): string {
        return this.expression;
    }
}
